#!/usr/bin/perl -T

use strict;
use MIME::Parser;
use File::Temp qw/ tempfile tempdir /;
use URI::URL;
use URI::Escape;
use Socket;

sub homedir(){
	$ENV{HOME}=~/(.*)/;
	$::homedir=$1;
}
use lib '/usr/local/lib';
use lib homedir."/.lib";
use lib '/usr/local/lib/mime_parse_scan';
use lib "$::homedir/.lib/mime_parse_scan";
use lib "$::homedir/.config";
use lib "$::homedir/.config/mime_parse_scan";
use BE::Lib;

my %maptype=("text/x-mail"=>"text/plain");
my @ignoretype=(("text/html","text/plain"));


#globals:

$::L=0;
$::L2=0;
$::text='';
%::links=();
%::tags=();

@::options=(
["","filtering:"],
["nodgout","skip files identified by dglist"],
["nocopy","do not copy new binaries to the current dir (*.bin)"],
["allcopy","copy all parts to the current dir (<filename>.part*)"],
["badurl","show bad urls, even if they are in the database"],
["allurl","show all urls, even if they are known non-spammers"],
[" ","(default is to show only new urls"],
["nonxdomain","skip urls if the domain does not exist"],
["spam","find spam and well-known (dglist) malware"],
["classify","detect abusereplies and delivery failures"],
["safe-classify","stop classifying if unsure"],
["headers","show headers"],
["allheaders","show all headers"],
["allparts","show all mime-parts"],
["procmail","-spam -onlyclassified -once"],

["","output options:"],
["L","list files without identified anomalities"],
["l","list files with identified anomalities"],
["onlyxxx[:yyy]","only output field xxx, subfield yyy (may be used multiple times)"],
["hidexxx","hide output field xxx (may be used multiple times)"],
["once","show only one block and only one diagnosis per kind"],
["html","html output"],
["script","script-friendly output"],
["classify-print","print lines not being matched by 'classify' to stderr"],

["",""],
["h","help"],
["help","help"]
);
%::options_implies=(
"safe-classify"  => ["-classify"],
"classify-print" => ["-classify"],
"procmail"       => ["-spam","-onlyclassified","-once"],
"allheaders"     => ["-headers"]
);

@::results=("file","mimetype","effectivetype","filename","size","copied_to",
            "type","classified","link","scanresult",
            "header","received","trusted_received","userpage",
            "unknown_tags","alt_sizes");

$::goodtag = qr/[aiubps]|font|h[r1-6]|br|li|body|html|head|div|strong|meta|span|
img|ul|t[rdh]|title|table|center|strike|em|style|map|area|(?:no)?script|
form|pre/xi;


%::filter=();
%::subfilter=();
@::hide=();
@::hide2=();

$::dowre=qr/Mon|Tue|Wed|Thu|Fri|Sat|Sun/i;
$::monre=qr/Jan|Feb|Mar|Apr|Jun|Jul|Aug|Sep|Oct|Nov|Dec/i;
$::datere=qr/(?:$::dowre,\s+)?           #dow
			 \d{1,2}\s+                                         #dd
			 $::monre\s+ #MM
			 \d{2,4}\s+                                         #yy(yy)
			 \d\d:                                              #hh
			 \d\d:                                              #mm
			 \d\d\s+                                            #ss
			 [+-]\d{4}                                          #tzone
			 (?i:\s+\((?:(?:[CM]E|J)ST|[CM]ET)\)|[-+]\d\d\d)?/x;
$::hostnamere=qr/[a-z](?:[-a-z0-9]*[a-z0-9])?/i;                #rfc1034 ~
$::rhostnamere=qr/[a-z0-9](?:[-a-z0-9_]*[a-z0-9])?/i;           #relaxed ~
$::domainre=qr/(?:$::rhostnamere\.)*$::hostnamere\.?/;
$::mailatom=qr/[-0-9a-z+!#$%&'*\/=?^_`{|}~]+/i;
$::localpartre=qr/"(?:[^"]|\\")"		#"
				 |$::mailatom(?:\.$::mailatom)*/x;
$::mailre=qr/$::localpartre\@$::domainre/;
$::ipre=qr/\d{1,3}(?:\.\d{1,3}){3}/;
$::menginere=qr/$::domainre(?:[-\ ][0-9.]+)?(?:\s+\#\d+)?
			   |(?:\d+(?:\.\d+)+
			    |CommuniGate\ Pro\ SMTP)(?:\ [-0-9.a-z]+)?
			   |\d+(?:\.\d+)+\/\d+(?:\.\d+)+/ix;
$::mtransportre=qr/[ae]?smtp|pop3|spamassassin|http|Microsoft SMTPSVC/i;

%::links_match=('a'=>'href', 'img'=>'src', 'form'=>'action', 'area'=>'href',
'iframe'=>'src');
$::html_parser_t=0;


sub max($@){
	my $v=shift;
	while(defined(my $x=shift)){
		if($x<$v){$v=$x}
	}
	return $v;
}

sub help()
{
	for my $a (@::options) {
		my ($b,$c)=@{$a};
		my $l=length($b);
		print "$b".((!$l||$l>=12)?"\n".' ' x 12:' ' x (12-$l))."$c\n";
		if($b eq "" && $c ne ""){
		print '            '.('~' x length($c))."\n";
		}
	}
	print "\nvalid field captions are:\n";
	for(my $i=0;$i<int((@::results+8)/9);$i++){
		print join(', ',@::results[$i*9 .. max($i*9+8,scalar @::results-1)]),"\n";
	}
	exit 0;
}

sub html($){
	(my $a=$_[0])=~s/([^-_+*\/a-z0-9.,;:\$\@!?\\'"=() \t])/'&#'.ord($1).";"/egi;#"
	return $a;
}

sub link_str{
	my ($tag,$url)=@_;
	if(%::filter){
		if(!$::filter{"link"}){return ""}
		if(grep($_ eq "link",@::hide2)){return ""};
		if(!$::L && !grep($_ eq "link",@::hide)){$::L=1}
	}else{
		if(grep($_ eq "link",@::hide,@::hide2)){return ""};
		$::L=1;
	}
	$url=~m~^(https?:/+)?([0-9a-z]+\@)?([^/\?:]+)~i;
	my $domain=$3;
	if($::opt{"once"}){push(@::hide2,"link")}
	if($::opt{"html"}){
		$url=~/^(.*?)(\?.*)?$/;
		my ($a,$b)=(html($1),html($2));
		return "<tr><th>link ($tag)</th><td>"
		."<a href=\"$a\">$a</a>$b"
		.(!($domain=~/^mailto:/i)?
		 " (<a href=\"http://$domain\">".html($domain)."</a>)":"")
		."</td><tr>\n";
	}elsif(keys(%::filter)==1){
		return "$url\n";
	}elsif($::opt{"script"}){
		return "link_$tag $url\n";
	}else{
		return sprintf("link_%-8s = %s\n",$tag,$url);
}	}

sub field_str($$;$%){
	my ($h,$e,$l,%h)=@_;
	
	if(!grep($_ eq $h,@::results)){
		print STDERR "unexpected output header $h\n";
	}
	
	if(%::filter){
		if(!$::filter{$h}){return ""}
		if(grep($_ eq $h,@::hide2)){return ""};
		if($h eq "header" && (my $f = $::subfilter{"header"})){
			if(keys(%{$f})==1){
				$e=~s,^(\S+:)\s*(.*),$2,;
			}else{
				$e=~m,^(\S+:)\s*(.*),;
			}
			if(!${$f}{lc($1)}){return ""}
		}
		if(!$::L && !grep($_ eq $h,@::hide,'file')){$::L=1}
	}else{
		if(grep($_ eq $h,@::hide,@::hide2)){return ""};
		if($l){$::L=1}
	}
	#print "fs $::L $h $e $l\n";
	
	if(keys(%h)){
		if(my $f = $::subfilter{$h}){
			my %f=%{$f};
			my %hh;
			for my $a (keys(%f)){if(defined(my $t=$h{$a})){$hh{$a}=$t}}
			%h=%hh;
		}
		if($::subfilter{$h} && (keys(%{$::subfilter{$h}})==1)){
			return join("",values(%h))."\n";
		}else{$e=	join("; ",map("$_ => $h{$_}",sort(keys(%h))))}
	}

	if($::opt{"once"}|| ($h eq "file")){push(@::hide2,$h)}
	
	if($::opt{"html"}){
		$e=html($e);
		if($e eq ""){$e="&nbsp;"}
		if(%h){
			$e="<table border=0>\n <tr><th>".
			join("</td></tr>\n <tr><th>",
			map("$_</th><td>".html($h{$_}),sort(keys(%h))))
			."</td></tr></table>";
		}elsif($h eq "header"){
			$e=~s,^(\S+:),<em>$1</em>,;
		}

		$h=~y/_/ /;
		return "<tr><th width='100'>$h</th><td>$e</td><tr>\n";
	}elsif($::opt{"script"}){
		if(%h){
			my $p;
			if(keys(%::filter)==1){
				if($::subfilter{$h} && (keys(%{$::subfilter{$h}})==1)){
					return join("",values(%h))."\n";
				}else{$p=""}
			}else{$p="$h "}
			return	join("",map("$p$_ $h{$_}\n",sort(keys(%h))),
			(!$::subfilter{$h}
			|| (keys(%{$::subfilter{$h}})!=1))?"${p}endofblock\n":"");
		} else {
			return "$h $e\n";
		}
	}elsif(keys(%::filter)==1){
		return "$e\n";
	}else{
		return sprintf("%s%-13s = %s\n",($h eq "file")?"\n":"",$h,$e);
}	}

# html parser

sub filterurl($) {
	my $url=shift;

	$url=~s~(https?://)?([^/?]+)\.([/?].*)?$~$1$2$3~i;
	$url=~s~(https?:/)([^/])~$1/$2~i;
	if(!$url=~/^[a-z0-9]+:/){$url='http://'.$url}
	$url=~m~^(https?:/+)?([0-9a-z]+\@)?([^/\?:]+)~i;
	my $domain=$3;
	my $flag=0;
	
	if($url=~/^cid:/){return 0}
	
	my $d=$domain;
	if($d=~/^\d+$/){
		$d=join('.',unpack("C4",pack("N",$d)));
	}
	while($d){
		if($flag=$::DOMAINLIST{$d}){last}
		($d=~s/^.*?\.//)||($d='');
	}
	
	if((!$flag||$flag==2) && ($url=~$::userpagere)){
		my $userpage=$+;
		$flag=$::USERPAGELIST{$userpage};
		if(!$::userpage){$::userpage=$userpage}
	}
	
	if(!$flag && $domain=~/[%_]/){$flag=2}
	if($::opt{"spam"}&&($flag==1)&&!$::L2) {
		if($::classify){$::classify="url $url";}
	}
	if(!$flag
	||($::opt{"allurl"} && ($flag!=3 || !$::opt{"nonxdomain"}))
	||($::opt{"badurl"}&&$flag==1)){
		return 1
	}
	return 0;
}

sub my_deescape($){
	my $url1=shift;
	my ($url,$redir)=url_deescape($url1,"redir","correct");
	if($redir&&$::opt{"spam"}&&!$::L2) {
		if($::classify){$::classify=join(' ',@$redir)}
	}
	return $url;
}

sub cb_start {
	my($tag, $attr) = @_;
	$::tags{lc($tag)}++;
#	$::tags2{lc($tag)}++;
	for my $k (keys(%{$attr})){
		if($::links_match{$tag} && ($::links_match{$tag} eq $k)){
			my $url=my_deescape(${$attr}{$k});
			if(filterurl($url)){$::links{$tag}{$url}=1}
}	}	}

sub cb_end {
	my $tag = $_[0];
	$tag=~s,^/,,;
	$::tags{lc($tag)}--;
#	$::tags2{lc($tag)}++;
}

sub check_IP_block($){
#	print "checking IP $_[0]\n";
	if(!($::classify ne "")){return}
	if(!($_[0]=~/\d+(\.\d+){3}/)) {return}
	my $na=unpack('N',inet_aton($_[0]));
	for my $w (@::IP_BLOCKS){
		my ($x,$y,$z)=@$w;
#		print "... $x $na $y $z\n";
		if(($na >= $x) && ($na <= $y)) {
#			print "foo\n";
			$::classify="IP_range_block ($_[0]): $z";
		}
	}
}

use HTML::Parser;
$::HTML_Parser=HTML::Parser->new(
  default_h => [sub { $::text.=shift }, 'dtext'],
  start_h   => [\&cb_start, 'tag, attr'],
  comment_h => [""],
  end_h     => [\&cb_end, 'tag']);

sub classify(@){

	map {
		chomp;
		s/[\x0d\x0a]+//g;
		s/[\0-\037]/ /g;s/\s+/ /g;
		s/^\s+//;s/\s+$//;s/\s+/ /g;
		while(s/^([-*\#=]+)(.*)\1$/$2/g || s/\[(.*)\]/$1/){s/^\s+//;s/\s+$//;s/\s+/ /g}
	} @_;

	my @l=@_;
	my @s=();
	while(defined ($_=shift(@l))){
		if(1...m/^$/){
			if(/^Auto-Submitted: auto-replied$/) {
				return "abusereply";
			}
			next;
		}
		
		if(/^This report relates to a message you sent with the following header fields:$/){
			shift(@l);
			while(defined ($_=shift(@l))){
				if(/^\s*$/){
					unshift(@l,$_);
					last;
				}
			}
			next;
		}
		
		if(  /^[-*=]*$/
		|| /^Content-Transfer-Encoding: (quoted-printable|8bit|binary)$/
		|| /^The original message was received at \w+, \d+ \w+ \d+ \d+:\d+:\d+ [-+]\d+ \([A-Z]+\)$/
		|| /^At \d+:\d+ \d+\.\d+\.\d+ [-+]\d+, you wrote:$/
		|| /^[-+a-zA-Z0-9\/.]+=*$/
		|| /^[-=_0-9]+$/
		|| /^from \[[0-9.]+\]$/
		|| /[^\0-\177]{4}|^(?:zn:|znA|HvBmC)$/
		|| /\bmail-abuse-bot\@7eggert.dyndns.org\b/
		|| /^_NextPart_\d+_[-+a-zA-Z0-9\/.]+/
		||m,^http://\S+$,
		|| /^$::ignorelinesre$/
		|| /^$::ignorelines_bolre/
		|| /^from [-.a-z0-9]+ (?:\d{1,3}\.){3}\d{1,3}$/
		|| /^On $::dowre, (?:\d+ )?$::monre .*, mail-abuse-bot wrote:$/
		|| /^\$B.*\(B$/
		) {
			next;
		} elsif((/^This /gi
		 && (/\Gis our automatic reply system to inform you that your email has/
		  || /\Gis an automatically generated response to your message/
		  || /\Gis an auto-reply to confirm recieving your email\./
		  || /\Gis our automatic reply system to here to apologize for all/
		  || /\Gan automatic response to confirm that we have received/
		  || /\Gmessage has been automatically generated in response to (your message|the)$/
		  || /\Gmessage was automatically generated in response to the message you sent,/
		  || (/\Gis an /g
		   && (/\Gautomat(ic|ed) reply\.?$/
		    || /\Gauto generated message to inform you that your message sent from /
		    || /\Gautomatic reply to the message "/ #"
		    || /\GAUTO-REPLY to acknowledge that your mail to the Abuse/
		    || /\Gautomated reply from the [A-Za-z.]+ abuse reporting list\.$/
		   ))
		 ))
		|| (/^(\* )?Thank you for /g
		 && (/\Gyour report. We will investigate this matter and take$/
		  || /\Gyour message to abuse\@integraonline.com\./
		  || /\Gyour email to BigPond abuse\. Your issue will be investigated by one of our$/
		  || /\Gyour e-mail\. We would like to inform you that it has been$/
		  || /\Gyour time. We are investigating this,/
		  || /\Gyour recent report to Sprint Acceptable Use Policy/
		  || /\Gyour mail regarding spam mail\./
		  || /\Ginforming us./
		  || /^We wish to thank you for your message in which you informed us about an/
		  || /\Gtaking the time to report this to [.a-z]+\@[-a-z0-9.]+$/
		  || /\Gtaking the time to submit your message regarding$/
		  || /\Gcontacting Level 3's Network Security Operations department\.$/
		  || /\Gcontacting the /g
		   && (/\GAbuse department at /
		    || /\GSBC Internet Services Policy department./)
		  || /\Gwriting to Adelphia's Internet Policy Enforcement Team\./
		  || /\Gsubmitting a request. A summary of your request is below:$/
		 ))
		|| /NOTE: This is an auto-responder\. Because of the/
		|| /\bThis message was sent automatically to reply to your inquiry\./
		|| /^Your message dated .* with subject /
		|| /^THIS IS AN AUTO(?:RESPONSE ONLY|MATED RESPONSE|-RESPONSE MESSAGE|-ACKNOWLEDGEMENT)\b/
		|| / has been successfully distributed to the ABUSE list /
		|| /^$::abusereplyre$/
		|| /^$::abusereply_bolre/
		|| /^(?:\d+\.)+\d+ is one of our user\.$/
		){
			return "abusereply";
		} elsif(
		   /^THIS IS A WARNING ONLY.  YOU DO NOT NEED TO RESEND YOUR MESSAGE.$/
		|| /^THIS IS A WARNING MESSAGE ONLY$/
		|| /^$::deliverywarningre$/
		|| /^$::deliverywarning_bolre/
		) {
			return "delivery_warning";
		} elsif(
		   /^Hi\. This is the qmail-send program at [-a-z0-9.]+\.$/
		|| /^This message was created automatically by mail delivery software\.$/
		|| /^I\'m sorry to have to inform you that the message returned$/
		|| /^VIRUS ALERT$/
		|| /^Dies ist eine automatisch erstellte Benachrichtigung \+APw-ber den Zustellstatus./
		|| /^The recipient was unavailable to take delivery of the message/
		|| /^Delivery to the following recipients failed\./
		|| /^$::deliveryfailurere$/
		|| /^$::deliveryfailure_bolre/
		) {
			return "delivery_failure";
		} elsif(
		   /^The user\(s\) account is temporarily over quota\.$/
		) {
			return "delivery_tempfailure";
		} elsif (
		   /^$::abusesuccessre$/
		|| /^$::abusesuccess_bolre/
			                
		){
			return "abusereply_success";
		} elsif (
		   /Emails sent to this address WILL NOT be read./
		){
			return "abusereply_denied";
		}
		if($::opt{"safe-classify"}){
			map{chomp}(@l);
			@s=(@s,$_,@l);@l=()
		}else{push(@s,$_)}
	}
	if(@s && $::opt{"classify-print"}){
		print ERR join("\n",@s),"\n";
#	print ERR '"',join("\"\n\"",@s),"\"\n";
	}
	return undef;
}

sub docsnyder($$){
	my($helo,$ip)=@_;
	if($::classify ne ''
	&& $::opt{"spam"}
	&& $ip){
		$ip=~/(\d+)\.(\d+)\.(\d+)\.(\d+)/;
		my $ips="$1\\W$2\\W$3\\W$4|$4\\W$3\\W$2\\W$1|$4\\W$1\\W$2\\W$3|".
		        sprintf("%3.3d\\W%3.3d\\W%3.3d\\W%3.3d", $1,$2,$3,$4)."|".
		        sprintf("%3.3d\\W%3.3d\\W%3.3d\\W%3.3d", $4,$3,$2,$1)."|".
		        sprintf("%2.2x%2.2x%2.2x%2.2x", $1,$2,$3,$4)."|".
		        sprintf("%2.2x%2.2x%2.2x%2.2x", $4,$3,$2,$1);
#		print "DS: $helo $ip $ips\n";
		return $helo=~/$ips/i;
	}
	return 0;
}



#spamfilters:
require "mime_parse_scan/vars_from_lists";
require "mime_parse_scan/usersettings.pl";

#init:

my $tempdir = tempdir( CLEANUP => 1 );
my ($fh, $tempfile) = tempfile( DIR => $tempdir );
delete $ENV{PATH};

my $parser = MIME::Parser->new;
$parser->output_to_core(1);
$parser->decode_headers(1);
#$parser->extract_uuencode(1);
$parser->output_to_core('ALL');

%::DOMAINLIST=();
{
	my $i=1;
	for my $F (($::DOMAINLIST_BAD, $::DOMAINLIST_GOOD, $::DOMAINLIST_NX)){
		open(DL,'<',$F) || die "open($F): $!";
		while(<DL>){
			chomp($_);
			/^$|^#/&&next;
			s/^\[(.*)\]$/$1/;
			$::DOMAINLIST{$_}=$i;
		}
		close(DL) || die "close($F): $!";
		$i++;
	}
}

@::IP_BLOCKS=();
{
	open(BL,'<',$::IP_BLOCKS) || die "open($::IP_BLOCKS): $!";
	while(<BL>){
		chomp($_);
		/^$|^#/&&next;
		my @F=split(/\s+/,$_,3);
		if(!($F[0]=~/\d+(\.\d+){3}/)
		|| !($F[1]=~/\d+(\.\d+){3}/)){next}
		my $b=unpack('N',inet_aton($F[0]));
		my $e=unpack('N',inet_aton($F[1]));
		push(@::IP_BLOCKS,[$b,$e,$F[2]]);
	}
	close(BL) || die "close($::IP_BLOCKS): $!";
}

%::USERPAGELIST=();
{
	my $i=1;
	open(DL,'<',$::USERPAGELIST) || die "open($::USERPAGELIST): $!";
	while(<DL>){
		chomp($_);
		/^$|^#/&&next;
		s/^\[(.*)\]$/$1/;
		$::USERPAGELIST{$_}=1;
	}
	close(DL) || die "close($::USERPAGELIST): $!";
}

open(DGLIST,'<',$::DIGESTLIST) || die "open($::DIGESTLIST:) $!";
@::DGLIST=<DGLIST>;
close(DGLIST) || die $!;
map(chomp($_),@::DGLIST);
@::DGLIST=grep(!/^$|^#/,@::DGLIST);

open(OLDSTDIN,'<&STDIN');
open(STDIN,'<','/dev/null');
open(ERR,'>&STDERR');

my @argv=@ARGV;
file: while(my $F = shift(@argv)) {
	my $message="";
	my $top;
	my $num=-1;
	my $to;
	$::L2=0;
	my $opts=1;

	if($opts && $F=~/^-+(\S+?)(?:=(\S+))?$/){
		my $o=$1;
		my $v=$2;
		if(!defined $v){$v=1}
		$::opt{$o}=$v;
		if($o=~/^h(elp)?$/){help}
		if($o eq "html"){
			print "<html><head><title>mime-parse-scan</title></head><body>"
			 ."<h1>Options</h1>".join(' ',@ARGV)."<h1>Results</h1>";
		}
		if($o=~/^only([-a-z_]+)(?::(\S+))?$/i){
			$::filter{$1}=1;
			if($2){$::subfilter{$1}{$2}=1}
		} elsif($o=~/^hide(\w+)$/){
			push(@::hide,$1);
		} else {
			if(my $a=$::options_implies{$o}){unshift(@argv,@{$a})}
			if(!grep($_ eq $o,map(${$_}[0],@::options))){
				print "unknown option <$o> value <$v>\n";
				die;
		}	}
		next;
	}
	
	if($F eq '-'){
		if(!open(F,'<&OLDSTDIN')){
			print ERR "open(stdin): $!\n";
			next;
		}
		if($ENV{"FILENO"}=~/^\d+$/){$F="mbox-$ENV{FILENO}"}
	} else {
		if(!open(F,'<',$F)){
			print ERR "open($F): $!\n";
			next;
		}
		$F=~s/\n/\\n/g;
	} 
	{
		my $l;
		my $h=1;
		if($F=~/\.(?:bsmtp|post)$/){
			while(!(<F>=~/^DATA\b/)){};
			while(!(($l=<F>)=~/^\.[\12\15]+$/))	{
				if($h){
					if(!$l=~/^\s|^[-a-z0-9]+:/i){
						if($l){$message.="\015\012"}
						$h=0;
				}	}
				$l=~s/^\.//;
				$message.=$l;
			}
		} else {
			while($l=<F>){
				if($h){
					if(!$l=~/^\s|^[-a-z0-9]+:/i){
						if($l){$message.="\015\012"}
						$h=0;
				}	}
				$l=~s/^\.//;
				$message.=$l;
	}	}	}
	$top = eval {$parser->parse_data($message)};
	close(F);

#	$top->dump_skeleton(\*STDOUT);

	my $ent;
	my $pts=0; # parts remaining in this multipart/*
	my $ptss;
	my @pts=();
	part:foreach $ent ($top->parts_DFS) {

		$num++;
		$::L=0;
		$::classify="@";
		$::userpage="";
		@::hide2=();
		$::text='';

		my %links=();

		my $out=field_str("file","$F $num");

		if($pts){
			if(--$pts && $ent->parts){
				$pts=0;
			}
			if(!$pts){@pts=()}
		}

		my $mimetype = $ent->mime_type();
		my $effectivetype = $ent->effective_type;
		$out.=field_str("mimetype","$mimetype");
		
		if($effectivetype ne $mimetype){
			$out.=field_str("effectivetype",$effectivetype,1);
			if($effectivetype eq "application/x-unparseable-multipart"
			&& ($::opt{"classify"}
			 || $::opt{"spam"})){
				$::classify='';
				$out.=field_str("classified","unparsable",1);
			}
		}

		#(my $p=join("\n",@{$ent->preamble||[]}))=~s/^\s+|\s+$//g;
		#(my $e=join("\n",@{$ent->epilogue||[]}))=~s/^\s+|\s+$//g;
		#print ("n ",scalar $ent->parts,"\n");
		#$p&&print ("p ",length($p)," $p\n");
		#$e&&print ("e ",length($e)," $e\n");

		if(!$::opt{"allcopy"}){
			my $h=$ent->head;
			my $rt=@::received_chain?1:0;
			my %lastR=("Mail-Host" => '...');
			if($h && ${$h}{"mail_hdr_list"}) {
				my @H=@{${$h}{"mail_hdr_list"}};
				for (@H) {
					chomp;
					s/\s*\n\s+/ /gs;
					s/^(\w)(\w*:)\s*/\u$1\L$2\E /;
					my %R=();
					if(/^Received:\s*(.*)/){
						$_=$1;
						if((/^\(qmail (?:\d+ )?invoked by (uid \d+|alias)\); $::datere$/
						 || /^\Qfrom localhost [127.0.0.1] by \E
						      $::domainre\Q with SpamAssassin (\E[^)]+\);\s+$::datere/x)
						&& !$::opt{"allheaders"}
						){next}
						if(/^from\ ($::domainre|$::ipre|\[$::ipre\]|\d+) #1 helo or rdns
						    (?:\ \(
						      (?:($::localpartre|)\@)? #2 ident-b
						      (?:($::domainre)\ )? #3 rdns
						      (?:\[($::ipre)\])?   #4 IP
						      (?:\ *\(([^)]+)\))?   # 5 comment
						      (?:\ *(?i:helo|ehlo)[=\ ]($::domainre|$::ipre))?  #6 helo
						      (?:\ *ident=($::localpartre))? #7 ident
						    \))?
						    (?:\ \(($::ipre)\))? #8 IP-b
						    (?:\ \[($::ipre)\])? # #9 IP-c
						    \s+by\s+($::domainre) (?:\ \($::menginere\))? #10 mhost
						    \ with\ ($::mtransportre) #11 mtransport
						    	(?:\ *\($::menginere\))? #void mengine
						    (?:\ id\ [-0-9a-zA-Z]+)? #void id
						    (?:\([^)]*\)\ )?
						    (?:\ for\ ($::mailre|<$::mailre>) (?:\ \(single-drop\))?)? #12 to:
						    ;\ ($::datere)$/x #13 date
						){
							my $d1_is_IP = 0;
							#print "here2 $1,$2,$3,$4,$5,$6,$7,$8,$9,$10,$11,$12,$13,$14\n";
							%R=("Source-IP"    => $4,
								"Comment1"     => $5,
								"Source-HELO"  => $6,
							    "Source-Ident" => $7,
							    "Mail-Host"    => $10,
							    "Mail-Protocol"=> $11,
							    "Mail-To"      => $12,
							    "Date"         => $13);
							if(!$4){
								if($8){$R{"Source-IP"}=$8}
								elsif($9){$R{"Source-IP"}=$9}
								elsif($1 && $6){$R{"Source-IP"}=$1; $d1_is_IP=1;}
							}
							if(!$6){
								$R{"Source-HELO"}=$1;
								$R{"Source-rDNS"}=$3;
							}elsif(!$d1_is_IP){
								$R{"Source-rDNS"}=$1;
							}
							if($d1_is_IP){
								if($R{"Source-IP"} =~ s/^\[($::ipre)\]$/$1/) {
								}elsif(!($R{"Source-IP"} =~ /^$::ipre$/)){
									#my $a=$R{"Source-IP"};
									delete($R{"Source-IP"});
								}
							}
						}elsif(/^from\ \[($::ipre)\]
						    (?:\ \(
						      (?:(?i:helo|ehlo)[=\ ]($::domainre|$::ipre|\w+))?
						      (?:\ *ident=($::localpartre))?
						    \))?
						    \ by\ ($::domainre) (?:\ \($::menginere\))?
						    \ (?:with|via)\ ($::mtransportre)
						    	(?:\ *\($::menginere\))?
						    (?:\ id\ [-0-9a-zA-Z]+)?
						    (?:\ for\ ($::mailre|<$::mailre>) (?:\ \(single-drop\))?)?
						    ;\ ($::datere)$/x
						){
							%R=("Source-IP"    => $1,
							    "Source-HELO"  => $2,
							    "Source-Ident" => $3,
							    "Mail-Host"    => $4,
							    "Mail-Protocol"=> $5,
							    "Mail-To"      => $6,
							    "Date"         => $7);
						}elsif(/^from\ ($::domainre)
						    \ \(
						      (?:(?:($::localpartre)\@)?($::domainre))?
						      (?:\ *\[($::ipre)\])?(?:\ *\(may\ be\ forged\))?
						      (?:\ *ident=($::localpartre))?
						    \)
						    \ by\ ($::domainre) (?:\ \($::menginere\))?
						    \ with\ ($::mtransportre)
						    	(?:\ *\($::menginere\))?
						    (?:\ id\ (?:[-0-9a-zA-Z]+|<$::mailre>))?
						    	(?:\s*\([^)]+\))?
						    (?:\ for\ ($::mailre|<$::mailre>) (?:\ \(single-drop\))?)?
						    ;\ ($::datere)$/x
						){
							%R=("Source-HELO"  => $1,
							    "Source-rDNS"  => $3,
							    "Source-IP"    => $4,
							    "Source-Ident" => $5,
							    "Mail-Host"    => $6,
							    "Mail-Protocol"=> $7,
							    "Mail-To"      => $8,
							    "Date"         => $9);
							if($2&&!$5){$R{"Source-Ident"}=$2}
						}elsif(/^from\ ($::localpartre)
						    \ by\ ($::domainre) (?:\ \($::hostnamere\))?
						    \ with\ local
						    	(?:\ *\($::menginere\))?
						    (?:\ id\ [-0-9a-zA-Z]+)?
						    (?:\ for\ ($::mailre|<$::mailre>) (?:\ \(single-drop\))?)?
						    ;\ ($::datere)$/x
						){
							%R=("Source-Ident" => $1,
							    "Mail-Host"    => $2,
							    "Mail-Protocol"=> "local",
							    "Mail-To"      => $3,
							    "Date"         => $4);
						}elsif(/^from(?:\ ($::domainre))?\ \[($::ipre)\]
						    (?:\ \(helo=($::domainre|$::ipre|[-0-9]+)\))?
						    \ by\ ($::domainre)
						    \ with\ ($::mtransportre)
						    	(?:\ *\($::menginere\))?
                            (?:\ id\ [-0-9a-zA-Z]+)?
                            (?:\ for\ ($::mailre)(?:\ \(single-drop\))?)?
                            ;\ ($::datere)$/x
						){
							%R=("Source-IP"    => $2,
							    "Source-HELO"  => $3,
							    "Mail-Host"    => $4,
							    "Mail-Protocol"=> $5,
							    "MAIL-TO"      => $6,
							    "Date"         => $7);
							if(!$3 && $1){$R{"Source-HELO"}=$1};
						}else{
							if($rt && $::opt{"headers"} && !$::opt{"allheaders"}
							&& $lastR{"Mail-Host"} ne '...'){
								delete($lastR{"trusted"});
								$out.=field_str("trusted_received",
												'dontshowthis',1,%lastR);
							}
							if($::opt{"spam"} && $lastR{"Source-IP"}){
								check_IP_block($lastR{"Source-IP"})
							}
							$rt=0;
							if($::opt{"headers"}){
								$out.=field_str("received","(raw) $_",1)
							}
							if(docsnyder($lastR{"Source-HELO"}, $lastR{"Source-IP"})){
								$::classify='';
								$out.=field_str("classified",
								"header received DocSnyder  $lastR{\"Source-HELO\"}",1);
							}
						}
						for my $a (sort(keys(%R))){if(!defined $R{$a}){delete($R{$a})}}
						if(%R){
							if($rt){
								$rt=0;
								loop_1: for my $c (@::received_chain){
									my %r=%R;
									for my $h (@{$c}) {
										for my $k (keys(%{$h})){
											if(!($r{$k}=~/^(?:${$h}{$k})$/)){next loop_1}
										}
										%r=%lastR;
									}
									$rt=1;
									last;
								}
								if($rt){
									$R{"trusted"}="yes";
								}elsif($lastR{"Mail-Host"} ne '...'){
									if($::opt{"headers"} && !$::opt{"allheaders"}){
										delete($lastR{"trusted"});
										$out.=field_str("trusted_received",
														'dontshowthis',1,%lastR);
									}
									if($::opt{"spam"} && $lastR{"Source-IP"}){
										check_IP_block($lastR{"Source-IP"})
									}
								}
								if(docsnyder($lastR{"Source-HELO"}, $lastR{"Source-IP"})){
									$::classify='';
									$out.=field_str("classified",
									"header received DocSnyder $lastR{\"Source-HELO\"} $lastR{\"Source-HELO\"}",1);
								}
							}
							if($::opt{"headers"}
							&& (!$rt || $::opt{"allheaders"}))
							{$out.=field_str("received",'dontshowthis',1,%R)}
							%lastR=%R;
						}else{
							# empty received - what to do?
						}
					}else{
						/^([-a-z]+):\s+(.*)/i;
						my ($a,$v)=(lc($1),$2);
						if($::opt{"allheaders"}
						|| ($::opt{"headers"}
						 && !($a=~m<^(
							  mime-version
							 |status
							 |x-(status
							    |mime-parse-scan
							    |mimeole
							    |msmail-priority
							    |priority
							    |resent-(by|for|to)
							    |spam-(flag|level)))$>x)))
						{$out.=field_str("header",$_,1)}
						if($num==0){
							if($a eq "to"){
								$to=$v;
							}
						}
						if(!$pts
						&& m,^Content-Type:\s*multipart/alternative\b,i){
							$pts=1+$ent->parts;
							$ptss=$num;
							if($pts==1
							&& $::classify ne ''
							&& ($::opt{"classify"}
							 || $::opt{"spam"})){
								$::classify='';
								$out.=field_str("classified","empty_multipart",1);
							}
						}
					}
				}
				if($::opt{"spam"}){shfilters:{
					while(@H){
						if($H[0]=~/([^:]+):\s+(.*)/){
							my $h=lc($1);
							if(my @r=grep(${$_}[1] eq $h,@::shfilters)){
								for(my $i=0;$i<@H;){
									if(lc($H[$i])=~/^\Q$h\E:\s+(.*?)\s*$/i){
										my $v=$1;
										splice(@H,$i,1);
										for(@r){
											my($s,$p,$n)=@{$_}[0,2,3];
											#print "h $h $s $v $p $n\n";
											#print "h^$h^$v^$p\n";
											#if($p && $v=~$p){
											#	print "p $v $p\n";
											#}
											if($::classify ne ''
											&& (($p == undef)
											 || ($v=~$p))
											&& (($n == undef)
											 ||!($v=~$n))){
												 $out.=field_str("classified",
												  "header $s $v",1);
												 $::classify='';
												 last shfilters;
											}
										}
									}else{$i++}
								}
							}else{
								@H=grep(!/^\Q$h\E:/i,@H);
							}
						}else{
							shift(@H);
						}
					}
				}}
				if($rt && $::opt{"headers"} && !$::opt{"allheaders"}
				&& $lastR{"Mail-Host"} ne '...'){
					delete($lastR{"trusted"});
					$out.=field_str("trusted_received",
									'dontshowthis',1,%lastR);
				}
				if($::opt{"spam"} && $lastR{"Source-IP"}){
					check_IP_block($lastR{"Source-IP"})
				}
				if(docsnyder($lastR{"Source-HELO"}, $lastR{"Source-IP"})){
					$::classify='';
					$out.=field_str("classified",
					"header received DocSnyder $lastR{\"Source-HELO\"} $lastR{\"Source-IP\"}",1);
				}
		}	}
		if (my $bodyh=$ent->bodyhandle) {
			
			if($::opt{"allcopy"}){
				$F=~m,(?:.*/)?(.*),;
				open(G,'>',"$1.part$num")||die "open $1.part$num";
				print G $bodyh->as_string;
				close(G);
				my $mimetype = $ent->mime_type();
				$out.=field_str("copied_to","$1.part$num",1);
			} else {
			
				my $filename = $ent->head->recommended_filename;;
				
				if($filename){
					if($filename =~ /^\[Virus entfernt\]\s/){next}
					$out.=field_str("filename","$filename",1);
				}
				
# removed: this covers nearly everything.
#				if(($filename&&!($filename=~/\.(gif|jpe?g|png)$/i))
#				|| ($mimetype=~m,
#				 ^(audio/x-wav
#				  |text/(html|plain)
#				  |image/.*
#				  |application/octet-stream)$,x)){
				{
					my $bstring;

					open(G,'>',$tempfile)||die "open $tempfile";
					#my $f=$ent->bodyhandle;
					#my $io = $f->open;
					#my $a;
					#while(($io->read($a,8192))>0){print G $a}
					print G ($bstring=$bodyh->as_string);
					close(G);

					my $type;
					{
						open(STDERR,'/dev/null');
						$type=`/usr/bin/file -i '$tempfile'`;
						open(STDERR,'>&ERR');
					}
					chomp($type);
					$type=~s<.*:\s+([-a-z]+/[-a-z0-9+]+)
					(,?\s+\S+)*
					(;(\s+[a-z]+=\S+)+)?$><$1>x;
					
					if(($type ne $mimetype)
					&& ($maptype{$type} ne $mimetype)
					&& !grep($_ eq ($mimetype,$type),@ignoretype))
					{$out.=field_str("type",$type,1)}

					$out.=field_str("size",length($bstring));
					if($pts){
						$pts[$pts]=length($bstring);
					}
					if($num==0 && !length($bstring) && $::opt{"classify"}
					&& $::classify ne ''){
						$out.=field_str("classified","empty",1);
						$::classify='';
					}

					my $scan;
					my $md;
					my $bydglist;
					if($type=~m,^(application/x-(dosexec|zip)
					            |text/html
					            |image/.*)$,x) {
					
						if(open(DGLIST,'-|',@::DGLIST_BIN,$tempfile)){
							$md=<DGLIST>;
							close(DGLIST);
							chomp($md);
							$md=~s;^\Q$tempfile\E\s+;;;

							for my $a (@::DGLIST){
								if($a=~/ \Q$md\E$/){
									if($::opt{nodgout}){next part}
									if($::opt{spam}&& $::classify ne ''){
										$out.=field_str("classified","malware $a",1);
										$::classify='';
										last;
									}
									($scan=$a)=~s/\s.*//;
									$scan.='(by_dglist)';
									$bydglist=1;
									last;
							}	}
						}else{
							print STDERR "open(-| ",join(" ",(@::DGLIST,$tempfile)),": $!\n";
						}

						if(!$scan && !$::opt{"spam"} && -e '/usr/local/bin/f-prot') {
							my @scan = `/usr/local/bin/f-prot -ai -collect -dumb -list '$tempfile'`;
							my $i=2;
							for (@scan) {
								if($_=~/^Switches:/){
									splice(@scan,0,$i);
									last;
								}
								$i++;
							}
							$scan=$scan[0];@scan=undef;
							chomp($scan);
							$scan=~s;.*\Q$tempfile\E(->.+)?\s*(  Infection:\s+)?;;;
							while($scan=~s;\s+\(\S+\);;){};
						}
					}
					
					if(!$scan){
						%::links=();
						if($mimetype eq "text/html"
						|| $type eq "text/html"
							# this is a non-infected html.
						|| ($type eq "text/plain")
						 && ($bstring=$bodyh->as_string=~m,</$::goodtag,i)){

							%::tags=();
							%::tags2=();
							$::HTML_Parser->parse_file($tempfile);

							my ($tag,$url);
							for my $tag (sort(keys(%::links))) {
								for my $url (sort(keys(%{$::links{$tag}}))){
									$out.=link_str($tag,$url);
									if($::opt{"spam"}
									&& $::classify eq '@') {
										if($tag eq "img"
										&& $url=~qr/\?.*\Q$to\E/)
										{
											$::classify="webbug $url";
											next;
										}
										if ($url =~ /^javascript:window\.external\.AddFavorite\(.*?\)?/) {
											$::classify="exploit: add bookmark $1";
							}	}	}	}
							for my $k (keys(%::tags)){
								if($k=~/^$::goodtag$/){delete($::tags{$k})}
							}
							my @p=grep($::tags{$_}> 0,keys(%::tags));
							my @n=grep($::tags{$_}< 0,keys(%::tags));
							my @b=grep($::tags{$_}==0,keys(%::tags));
							%::tags=();
							if(scalar @p+scalar @n+scalar @b){
								$out.=field_str("unknown_tags",
								scalar @p." ".scalar @n." ".scalar @b." ".
								join(', ',@p)."; ".join(', ',@n)."; ".
								join(', ',@b),1);
								if($::opt{"spam"} && ($::classify eq '@')){
									if((@p>4)
									&& ((abs(scalar @p-@n)<3)
									 || (@b<3))){
										$::classify=
										"fake_html_tags ".join(', ',@p)."; ".
										join(', ',@n)."; ".join(', ',@b);
										$::classify=~y/a-zA-Z0-9_ ,;=/?/c;
							}	}	}
							if($::opt{"classify"}&&($::classify eq "@")){
								my $fh=$ent->open("r");
								if(my $s=classify(<$fh>)){$::classify=$s}
								$fh->close;
							}
						} elsif($mimetype=~m,^text/plain$,) {
							if($::opt{"classify"}&&($::classify eq "@")){
								my $fh=$ent->open("r");
								if(my $s=classify(<$fh>)){$::classify=$s}
								$fh->close;
							}

						} elsif(!($type=~m,image/(gif|jpeg),)){
							$scan='unknown';
						}

					}
					
					if($scan
					&& (!keys(%::filter)||$::filter{"copied_to"}||$::filter{scanresult})
					&& !($type eq "application/x-empty")) {
						if(!$md) {
							if(open(DGLIST,'-|',@::DGLIST_BIN,$tempfile)){
								$md=<DGLIST>;
								close(DGLIST);
								chomp($md);
								$md=~s;^\Q$tempfile\E\s+;;;
							}else{
								print ERR "open(-| ",join(" ",(@::DGLIST,$tempfile)),": $!\n";
							}
						}

						$out.=field_str("scanresult","$scan $md",1);
						if(!$bydglist
						&& !$::opt{"l"}
						&& !$::opt{"L"}
						&& !$::opt{"spam"}
						&& !$::opt{"nocopy"}
						&&(!keys(%::filter)
						 ||($::filter{"copied_to"}))
						){
							(my $name=$md)=~s/\b\S+://g;
							$name=~s/ /./g;
							$name.=".bin";
							if(! (-e "$name")){
								if($name=~/^([.a-z0-9]+)$/){
									$name=$1;
									system("/bin/cp","-i","--",$tempfile,$name);
							}	}
							if( -e "$name") {
								$out.=field_str("copied_to",$name,1);
			}	}	}	}	}

			if(!$::text){$::text=$bodyh->as_string};

			{
				#print ERR $::text;
				#print ERR "\n------------------\n";
				if($pts){
					$pts[$pts]=length($::text);
				}
				$::text=~s/\xa0/ /g;
				while($::text=~m,(^|[<="\s])(https?:/+[^\s">]+),gi)
				{$links{my_deescape($2)}=1}
				while($::text=~m,\s(www\.[^\s"]+),gi)
				{$links{my_deescape("http://$1")}=1}
				while($::text=~m,^(www\.[^\s"]+),gmi)
				{$links{my_deescape("http://$1")}=1}

				for my $i (keys(%links)){
					if(!filterurl($i)){
						delete $links{$i};
				}	}

				for my $url (keys(%links)) {
					$out.=link_str('t',$url);
				}
			}

		}

		if($::userpage){
			$out.=field_str("userpage",$::userpage,1);
		}
		if($::classify && ($::classify ne "@")){
			$out.=field_str("classified",$::classify,1);
		}
		if($::L||$::opt{"allparts"}){
			$::L2|=$::L;
			if(!$::opt{"L"}
			&& !$::opt{"l"}
			&& $out ne ""){
				if($::opt{"html"}){print "<table border=1 width='100%'>\n"}
				print $out;
				if($::opt{"html"}){print("</table><br>&nbsp;<br>")}
			}
			if($::opt{"once"}){last part}
		}


		if($pts==1 && @pts){
			# last in this multipart/*
			shift(@pts);
			@pts=map(defined($_)?$_:"undef",reverse(@pts));
			@::hide2=();
			$::L=0;
			$out=field_str("file","$F $ptss");
			
			$out.=field_str("alt_sizes",join(', ',@pts),1);
			if($::opt{"spam"} && ($::classify eq '@')){
				if(!grep($_ eq "undef",@pts)){
					if(@pts==1&& $::classify ne ''){
						$out.=field_str("classified",
						 sprintf("alt_sizes_one %d",$pts[0]||0),1);
						$::classify='';
					}elsif(@pts==2){
						my $d;
						if($pts[0]&&$pts[1]){
							$d=$pts[0]/$pts[1];
							if($d<1){$d=1/$d}
							if($d>2&& $::classify ne ''){
								$out.=field_str("classified",
								 sprintf("alt_sizes_mismatch %.2f",$d),1);
								$::classify='';
							}
						}else{
							if($pts[0]!=$pts[1]&& $::classify ne ''){
								$out.=field_str("classified","alt_sizes_zero",1);
								$::classify='';
							}
						}
					}
				}	
			}
			if($::L||$::opt{"allparts"}){
				$::L2|=$::L;
				if(!$::opt{"L"}
				&& !$::opt{"l"}
				&& $out ne ""){
					if($::opt{"html"}){print "<table border=1 width='100%'>\n"}
					print $out;
					if($::opt{"html"}){print("</table><br>&nbsp;<br>")}
				}
				if($::opt{"once"}){last part}
			}
		}
	}
	if(($::opt{"L"}&&!$::L2)||($::opt{"l"}&&$::L2)){print $F." \n"}
	close(F);
}
if($::opt{"html"}){print("</body></html>\n")}
